package com.lutao.cloud.oil.runner;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.json.JSONUtil;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.lutao.cloud.oil.pojo.entity.TagOnlineRate;
import com.lutao.cloud.oil.service.*;
import com.lutao.entity.*;
import lombok.extern.slf4j.Slf4j;
import org.springframework.data.redis.core.BoundHashOperations;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * 在线率定时执行器
 *
 * @author yx
 * @date 2024/04/08
 */
@Component
@Slf4j
public class OnlineRatingExecutor {

    @Resource(name = "redisTemplateAlias")
    private RedisTemplate redisTemplate;

    @Resource
    private PlatformInfoService platformInfoService;

    @Resource
    private TagOnlineRateService tagOnlineRateService;

    @Resource
    private CollectLogService collectLogService;

    @Resource
    private FireParametersTagService fireParametersTagService;

    @Resource
    private GasParametersTagService gasParametersTagService;

    @Resource
    private HazardousGasParametersTagService hazardousGasParametersTagService;

    @Resource
    private ProcessParameterTagService processParameterTagService;

    private static final String LOCK_KEY = "scheduled_task_lock:OnlineRatingExecutor";

    private boolean acquireLock() {
        // 尝试获取锁，设置超时时间为 1 分钟
        return Boolean.TRUE.equals(redisTemplate.opsForValue().setIfAbsent(LOCK_KEY, "locked", 1, TimeUnit.MINUTES));
    }

    private void releaseLock() {
        // 释放锁
        redisTemplate.delete(LOCK_KEY);
    }

    /**
     * 0点启动初始化 指标信息
     */
    @Scheduled(cron = "0 0 0 * * ?")
    public void initProcess() {
        if (acquireLock()) {
            try {
                //存在的平台列表
                List<PlatformInfo> nowPlatformInfos = platformInfoService.list(Wrappers.lambdaQuery(PlatformInfo.class).eq(PlatformInfo::getDeleteFlag, 0));
                List<String> platformCodes = nowPlatformInfos.stream().map(PlatformInfo::getPlatformCode).collect(Collectors.toList());
                String currentTime = LocalDateTime.now().toString();
                List<FireParametersTag> fireParametersTags = fireParametersTagService
                        .list(Wrappers.lambdaQuery(FireParametersTag.class)
                                .isNotNull(FireParametersTag::getTagCode)
                                .isNotNull(FireParametersTag::getPlatformCode)
                                .isNotNull(FireParametersTag::getAppId)
                                .eq(FireParametersTag::getDelFlag,"0")
                                .eq(FireParametersTag::getStatus, "0")
                                // 只处理存在的平台
                                .in(FireParametersTag::getPlatformCode,platformCodes));

                for (FireParametersTag fireParametersTag : fireParametersTags) {
                    BoundHashOperations boundHashOperations = redisTemplate.boundHashOps(LocalDate.now() + ":"
                            + PerceptionMonitorFireInfo.TABLE_NAME
                            + ":" + fireParametersTag.getAppId()
                            + ":" + fireParametersTag.getPlatformCode()
                            + ":" + fireParametersTag.getTagCode());
                    boundHashOperations.put("should_accept_count", 0);
                    boundHashOperations.put("received_count", 0);
                    boundHashOperations.put("begin_time", currentTime);
                    boundHashOperations.put("enable", 0);
                    boundHashOperations.put("resource", PerceptionMonitorFireInfo.TABLE_NAME);
                    boundHashOperations.put("tagInfo", JSONUtil.toJsonStr(fireParametersTag));

                    boundHashOperations.expire(48, TimeUnit.HOURS);
                }

                List<GasParametersTag> gasParametersTags = gasParametersTagService
                        .list(Wrappers.lambdaQuery(GasParametersTag.class)
                                .isNotNull(GasParametersTag::getTagCode)
                                .isNotNull(GasParametersTag::getPlatformCode)
                                .isNotNull(GasParametersTag::getAppId)
                                .eq(GasParametersTag::getStatus, "0")
                                .eq(GasParametersTag::getDelFlag,"0")
                                .in(GasParametersTag::getPlatformCode,platformCodes));
                for (GasParametersTag gasParametersTag : gasParametersTags) {
                    BoundHashOperations boundHashOperations = redisTemplate.boundHashOps(LocalDate.now()
                            + ":" + PerceptionMonitorGasInfo.TABLE_NAME
                            + ":" + gasParametersTag.getAppId()
                            + ":" + gasParametersTag.getPlatformCode()
                            + ":" + gasParametersTag.getTagCode());
                    boundHashOperations.put("should_accept_count", 0);
                    boundHashOperations.put("received_count", 0);
                    boundHashOperations.put("begin_time", currentTime);
                    boundHashOperations.put("enable", 0);
                    boundHashOperations.put("resource", PerceptionMonitorGasInfo.TABLE_NAME);
                    boundHashOperations.put("tagInfo", JSONUtil.toJsonStr(gasParametersTag));
                    boundHashOperations.expire(48, TimeUnit.HOURS);
                }

                List<HazardousGasParametersTag> hazardousGasParametersTags = hazardousGasParametersTagService
                        .list(Wrappers.lambdaQuery(HazardousGasParametersTag.class)
                                .isNotNull(HazardousGasParametersTag::getTagCode)
                                .isNotNull(HazardousGasParametersTag::getPlatformCode)
                                .isNotNull(HazardousGasParametersTag::getAppId)
                                .eq(HazardousGasParametersTag::getDelFlag,"0")
                                .in(HazardousGasParametersTag::getPlatformCode,platformCodes)
                                .eq(HazardousGasParametersTag::getStatus, "0"));

                for (HazardousGasParametersTag hazardousGasParametersTag : hazardousGasParametersTags) {
                    BoundHashOperations boundHashOperations = redisTemplate.boundHashOps(LocalDate.now()
                            + ":" + PerceptionMonitorHazardousGasInfo.TABLE_NAME
                            + ":" + hazardousGasParametersTag.getAppId()
                            + ":" + hazardousGasParametersTag.getPlatformCode()
                            + ":" + hazardousGasParametersTag.getTagCode());
                    boundHashOperations.put("should_accept_count", 0);
                    boundHashOperations.put("received_count", 0);
                    boundHashOperations.put("begin_time", currentTime);
                    boundHashOperations.put("enable", 0);
                    boundHashOperations.put("resource", PerceptionMonitorHazardousGasInfo.TABLE_NAME);
                    boundHashOperations.put("tagInfo", JSONUtil.toJsonStr(hazardousGasParametersTag));
                    boundHashOperations.expire(48, TimeUnit.HOURS);
                }

                List<ProcessParameterTag> processParameterTags = processParameterTagService
                        .list(Wrappers.lambdaQuery(ProcessParameterTag.class)
                                .isNotNull(ProcessParameterTag::getTagCode)
                                .isNotNull(ProcessParameterTag::getPlatformCode)
                                .isNotNull(ProcessParameterTag::getAppId)
                                .in(ProcessParameterTag::getPlatformCode,platformCodes)
                                .eq(ProcessParameterTag::getDelFlag,"0")
                                .eq(ProcessParameterTag::getStatus, "0"));
                for (ProcessParameterTag processParameterTag : processParameterTags) {
                    BoundHashOperations boundHashOperations = redisTemplate.boundHashOps(LocalDate.now()
                            + ":" + PerceptionMonitorInfo.TABLE_NAME
                            + ":" + processParameterTag.getAppId()
                            + ":" + processParameterTag.getPlatformCode()
                            + ":" + processParameterTag.getTagCode());
                    boundHashOperations.put("should_accept_count", 0);
                    boundHashOperations.put("received_count", 0);
                    boundHashOperations.put("begin_time", currentTime);
                    boundHashOperations.put("enable", 0);
                    boundHashOperations.put("resource", PerceptionMonitorInfo.TABLE_NAME);
                    boundHashOperations.put("tagInfo", JSONUtil.toJsonStr(processParameterTag));
                    boundHashOperations.expire(48, TimeUnit.HOURS);
                }
            } finally {
                releaseLock();
            }
        }


    }


    /**
     * 在线率定时任务
     */
    @Transactional
    @Scheduled(cron = "0 0 1 * * ?")
    public void execute() {
        if (acquireLock()) {
            try {
                long startTime = System.currentTimeMillis();

                Set<String> redisKeys = redisTemplate.keys(LocalDate.now().minusDays(1) + "*");
                List<PlatformInfo> platformInfos = platformInfoService.list(Wrappers.lambdaQuery(PlatformInfo.class).eq(PlatformInfo::getDeleteFlag,0));

                Map<String, PlatformInfo> platformMap = new HashMap<>();
                for (PlatformInfo platformInfo : platformInfos) {
                    platformMap.put(platformInfo.getPlatformCode(), platformInfo);
                }
                if (redisKeys == null) {
                    return;
                }
                for (String key : redisKeys) {
                    processKey(key, platformMap);
                }
                long endTime = System.currentTimeMillis();
                long executionTime = endTime - startTime; //milliseconds
                String executionTimeStr = String.format("%d milliseconds", executionTime);
                CollectLog collectLog = new CollectLog();
                collectLog.setCommand("tag_online_rate");
                collectLog.setData(executionTimeStr);
                collectLogService.save(collectLog);

            } catch (Exception e) {
                CollectLog collectLog = new CollectLog();
                collectLog.setCommand("error");
                collectLog.setData("tag_online_rate==error==" + e.getMessage());
                e.printStackTrace();
                collectLogService.save(collectLog);
            } finally {
                releaseLock();
            }
        }
    }

    public void processKey(String key, Map<String, PlatformInfo> platformMap) {

        try {
            log.info("key==" + key);
            BoundHashOperations boundHashOperations = redisTemplate.boundHashOps(key);
            String resource = (String) boundHashOperations.get("resource");
            Integer receivedCount = Optional.ofNullable((Integer) boundHashOperations.get("received_count")).orElse(0);
            String entity = (String) boundHashOperations.get("entity");
            String tagInfo = (String) boundHashOperations.get("tagInfo");
            LocalDateTime beginTime = parseDateTime((String) boundHashOperations.get("begin_time"));
            LocalDateTime endTime = parseDateTime((String) boundHashOperations.get("end_time"));
            Integer shouldAcceptCount = Optional.ofNullable((Integer) boundHashOperations.get("should_accept_count")).orElse(0);

            TagOnlineRate tagOnlineRate = new TagOnlineRate();
            switch (Objects.requireNonNull(resource)) {
                case "perception_monitor_fire_info":
                    PerceptionMonitorFireInfo fireInfo = JSONUtil.toBean(entity, PerceptionMonitorFireInfo.class); // 将JSON字符串转换为Java对象
                    if (fireInfo.getTagCode() == null){
                        FireParametersTag fireParametersTag = JSONUtil.toBean(tagInfo, FireParametersTag.class);
                        fireInfo = new PerceptionMonitorFireInfo();
                        BeanUtil.copyProperties(fireParametersTag, fireInfo);
                    }
                    tagOnlineRate = processFireInfo(fireInfo, tagOnlineRate, platformMap, receivedCount, beginTime, endTime, resource, shouldAcceptCount);
                    break;
                case "perception_monitor_gas_info":
                    PerceptionMonitorGasInfo gasInfo = JSONUtil.toBean(entity, PerceptionMonitorGasInfo.class); // 将JSON字符串转换为Java对象
                    if (gasInfo.getTagCode() == null){
                        GasParametersTag gasParametersTag = JSONUtil.toBean(tagInfo, GasParametersTag.class);
                        gasInfo = new PerceptionMonitorGasInfo();
                        BeanUtil.copyProperties(gasParametersTag, gasInfo);
                    }

                    tagOnlineRate = processGasInfo(gasInfo, tagOnlineRate, platformMap, receivedCount, beginTime, endTime, resource, shouldAcceptCount);
                    break;
                case "perception_monitor_hazardous_gas_info":
                    PerceptionMonitorHazardousGasInfo hazardousGasInfo = JSONUtil.toBean(entity, PerceptionMonitorHazardousGasInfo.class); // 将JSON字符串转换为Java对象
                    if (hazardousGasInfo.getTagCode() == null){
                        HazardousGasParametersTag hazardousGasParametersTag = JSONUtil.toBean(tagInfo, HazardousGasParametersTag.class);
                        hazardousGasInfo = new PerceptionMonitorHazardousGasInfo();
                        BeanUtil.copyProperties(hazardousGasParametersTag, hazardousGasInfo);
                    }
                    tagOnlineRate = processHazardousGasInfo(hazardousGasInfo, tagOnlineRate, platformMap, receivedCount, beginTime, endTime, resource, shouldAcceptCount);
                    break;
                case "perception_monitor_info":
                    PerceptionMonitorInfo monitorInfo = JSONUtil.toBean(entity, PerceptionMonitorInfo.class); // 将JSON字符串转换为Java对象
                    if (monitorInfo.getTagCode() == null){
                        ProcessParameterTag processParameterTag = JSONUtil.toBean(tagInfo, ProcessParameterTag.class);
                        monitorInfo = new PerceptionMonitorInfo();
                        BeanUtil.copyProperties(processParameterTag, monitorInfo);
                    }
                    tagOnlineRate = processMonitorInfo(monitorInfo, tagOnlineRate, platformMap, receivedCount, beginTime, endTime, resource, shouldAcceptCount);
                    break;
                default:
                    break;
            }
            if (tagOnlineRate != null){
                if (tagOnlineRate.getPlatformCode() != null){
                    tagOnlineRateService.save(tagOnlineRate);
                }
            }
        }catch (Exception e){
            CollectLog collectLog = new CollectLog();
            collectLog.setCommand("error");
            collectLog.setData("tag_online_rate==error==" + e.getMessage()+key);
            e.printStackTrace();
            collectLogService.save(collectLog);
        }
    }

    private LocalDateTime parseDateTime(String dateTimeStr) {
        List<String> formats = Arrays.asList(
                "yyyy-MM-dd'T'HH:mm:ss",
                "yyyy-MM-dd'T'HH:mm:ss.SSS"
        );
        LocalDateTime dateTime = null;
        if (dateTimeStr != null) {
            for (String format : formats) {
                try {
                    DateTimeFormatter formatter = DateTimeFormatter.ofPattern(format);
                    dateTime = LocalDateTime.parse(dateTimeStr, formatter);
                    break;
                } catch (DateTimeParseException e) {
                    // Ignore and try the next format
                }
            }
        }
        return dateTime;
    }

    private TagOnlineRate processFireInfo(PerceptionMonitorFireInfo fireInfo, TagOnlineRate tagOnlineRate, Map<String, PlatformInfo> platformMap, Integer receivedCount, LocalDateTime beginTime, LocalDateTime endTime, String resource, Integer shouldAcceptCount) {
        if (fireInfo != null) {
            return fillTagOnlineRate(fireInfo.getAppId(), fireInfo.getPlatformCode(), fireInfo.getTagId(), fireInfo.getTagCode(), fireInfo.getTagName(), "6", tagOnlineRate, platformMap, receivedCount, beginTime, endTime, resource, shouldAcceptCount);
        } else return null;
    }

    private TagOnlineRate processGasInfo(PerceptionMonitorGasInfo gasInfo, TagOnlineRate tagOnlineRate, Map<String, PlatformInfo> platformMap, Integer receivedCount, LocalDateTime beginTime, LocalDateTime endTime, String resource, Integer shouldAcceptCount) {
        if (gasInfo != null) {
            return fillTagOnlineRate(gasInfo.getAppId(), gasInfo.getPlatformCode(), gasInfo.getTagId(), gasInfo.getTagCode(), gasInfo.getTagName(), "7", tagOnlineRate, platformMap, receivedCount, beginTime, endTime, resource, shouldAcceptCount);
        } else return null;
    }

    private TagOnlineRate processHazardousGasInfo(PerceptionMonitorHazardousGasInfo hazardousGasInfo, TagOnlineRate tagOnlineRate, Map<String, PlatformInfo> platformMap, Integer receivedCount, LocalDateTime beginTime, LocalDateTime endTime, String resource, Integer shouldAcceptCount) {
        if (hazardousGasInfo != null) {
            return fillTagOnlineRate(hazardousGasInfo.getAppId(), hazardousGasInfo.getPlatformCode(), hazardousGasInfo.getTagId(), hazardousGasInfo.getTagCode(), hazardousGasInfo.getTagName(), "8", tagOnlineRate, platformMap, receivedCount, beginTime, endTime, resource, shouldAcceptCount);
        } else return null;
    }

    private TagOnlineRate processMonitorInfo(PerceptionMonitorInfo monitorInfo, TagOnlineRate tagOnlineRate, Map<String, PlatformInfo> platformMap, Integer receivedCount, LocalDateTime beginTime, LocalDateTime endTime, String resource, Integer shouldAcceptCount) {
        if (monitorInfo != null) {
            ProcessParameterTag processParameterTag = processParameterTagService.getOne(Wrappers.lambdaQuery(ProcessParameterTag.class).eq(ProcessParameterTag::getTagCode, monitorInfo.getTagCode()).eq(ProcessParameterTag::getAppId, monitorInfo.getAppId()));
            String tagType = "";
            if (processParameterTag != null) {
                tagType = processParameterTag.getTagType();
            }
            return fillTagOnlineRate(monitorInfo.getAppId(), monitorInfo.getPlatformCode(), monitorInfo.getTagId(), monitorInfo.getTagCode(), monitorInfo.getTagName(), tagType, tagOnlineRate, platformMap, receivedCount, beginTime, endTime, resource, shouldAcceptCount);
        } else return null;
    }


    private TagOnlineRate fillTagOnlineRate(String appId, String platformCode, String tagId, String tagCode, String tagName, String tagType, TagOnlineRate tagOnlineRate, Map<String, PlatformInfo> platformMap, Integer receivedCount, LocalDateTime beginTime, LocalDateTime endTime, String resource, Integer shouldAcceptCount) {
        tagOnlineRate.setAppId(appId);
        tagOnlineRate.setPlatformCode(platformCode);
        PlatformInfo platformInfo = platformMap.get(platformCode);
        if (platformInfo != null) {
            //说明没有查到平台信息 直接return
            tagOnlineRate.setOilGasName(platformMap.get(platformCode).getOilGasName());
            tagOnlineRate.setPlatformName(platformMap.get(platformCode).getPlatformName());
            tagOnlineRate.setOperatingCompany(platformMap.get(platformCode).getOperatingCompany());
            tagOnlineRate.setTagId(tagId);
            tagOnlineRate.setTagCode(tagCode);
            tagOnlineRate.setTagName(tagName);
            tagOnlineRate.setTagType(tagType);
            tagOnlineRate.setBeginTime(beginTime);
            tagOnlineRate.setEndTime(endTime);
            tagOnlineRate.setReceiveCount(receivedCount);
            tagOnlineRate.setShouldReceiveCount(shouldAcceptCount);
            if (shouldAcceptCount == 0) {
                shouldAcceptCount = 143;
            }
            tagOnlineRate.setOnlineRate(receivedCount != null ? String.valueOf((float) receivedCount / shouldAcceptCount) : "0");
            tagOnlineRate.setResourceTable(resource);
            tagOnlineRate.setCreateTime(LocalDateTime.now());
            tagOnlineRate.setUpdateTime(LocalDateTime.now());
            tagOnlineRate.setDataDate(LocalDate.now().minusDays(1).toString());
            return tagOnlineRate;
        }else {
            return null;
        }

    }

}
